# -*- coding: utf-8 -*-
"""
Helper functions for multiprocessing in snr generation
"""

import numpy as np
import bilby

from ..numba import noise_weighted_inner_product


def noise_weighted_inner_prod_h_inner_h(params):
    """
    Probaility of detection of GW for the given sensitivity of the detectors

    Parameters
    ----------
    params : list
        list of parameters for the inner product calculation
        List contains: \n
        params[0] : float
            mass_1
        params[1] : float
            mass_2
        params[2] : float
            luminosity_distance
        params[3] : float
            theta_jn
        params[4] : float
            psi
        params[5] : float
            phase
        params[6] : float
            ra
        params[7] : float
            dec
        params[8] : float
            geocent_time
        params[9] : float
            a_1
        params[10] : float
            a_2
        params[11] : float
            tilt_1
        params[12] : float
            tilt_2
        params[13] : float
            phi_12
        params[14] : float
            phi_jl
        params[15] : float
            lambda_1
        params[16] : float
            lambda_2
        params[17] : float
            eccentricity
        params[18] : float
            approximant
        params[19] : float
            f_min
        params[20] : float
            f_ref
        params[21] : float
            duration
        params[22] : float
            sampling_frequency
        params[23] : int
            index tracker
        params[24] : list
            list of psds for each detector
        params[25] : str
            frequency_domain_source_model name
        

    Returns
    -------
    SNRs_list : list
        contains opt_snr for each detector and net_opt_snr
    params[22] : int
        index tracker
    """

    bilby.core.utils.logger.disabled = True
    np.random.seed(88170235)
    parameters = {
        "mass_1": params[0],
        "mass_2": params[1],
        "luminosity_distance": params[2],
        "theta_jn": params[3],
        "psi": params[4],
        "phase": params[5],
        "geocent_time": params[8],
        "ra": params[6],
        "dec": params[7],
        "a_1": params[9],
        "a_2": params[10],
        "tilt_1": params[11],
        "tilt_2": params[12],
        "phi_12": params[13],
        "phi_jl": params[14],
        "lambda_1": params[15],
        "lambda_2": params[16],
        "eccentricity": params[17],
    }

    # print('eccentricity', params[17])
    # print('frequency_domain_source_model', params[25])

    waveform_arguments = dict(
        waveform_approximant=params[18],
        reference_frequency=params[20],
        minimum_frequency=params[19],
    )

    waveform_generator = bilby.gw.WaveformGenerator(
        duration=params[21],
        sampling_frequency=params[22],
        frequency_domain_source_model=getattr(bilby.gw.source, params[25]),
        waveform_arguments=waveform_arguments,
    )
    polas = waveform_generator.frequency_domain_strain(parameters=parameters)

    # h = F+.h+ + Fx.hx
    # <h|h> = F+^2<h+,h+> + Fx^2<hx,hx> + F+Fx 2<h+,hx>
    # <h|h> = F+^2<h+,h+> + Fx^2<hx,hx>, if h+ and hx are orthogonal
    # F+^2 and Fx^2 will be added later
    hp_inner_hp_list = []
    hc_inner_hc_list = []
    # list_of_detectors = params[26:].tolist()
    psds_objects = params[24]
    for idx in range(len(psds_objects)):
    # for idx, det in enumerate(list_of_detectors):

        # need to compute the inner product for
        p_array = psds_objects[idx][2](waveform_generator.frequency_array)
        # p_array =  cubic_spline_interpolator(xnew_array=waveform_generator.frequency_array, coefficients=psds_objects[idx][2], x=psds_objects[idx][0])

        idx2 = (p_array != 0.0) & (p_array != np.inf)
        hp_inner_hp = noise_weighted_inner_product(
            polas["plus"][idx2],
            polas["plus"][idx2],
            p_array[idx2],
            waveform_generator.duration,
        )
        hc_inner_hc = noise_weighted_inner_product(
            polas["cross"][idx2],
            polas["cross"][idx2],
            p_array[idx2],
            waveform_generator.duration,
        )

        hp_inner_hp_list.append(hp_inner_hp)
        hc_inner_hc_list.append(hc_inner_hc)

    return (hp_inner_hp_list, hc_inner_hc_list, params[23])

def noise_weighted_inner_prod_d_inner_h(params):
    """
    Probaility of detection of GW for the given sensitivity of the detectors

    Parameters
    ----------
    params : list
        list of parameters for the inner product calculation
        List contains: \n
        params[0] : float
            mass_1
        params[1] : float
            mass_2
        params[2] : float
            luminosity_distance
        params[3] : float
            theta_jn
        params[4] : float
            psi
        params[5] : float
            phase
        params[6] : float
            ra
        params[7] : float
            dec
        params[8] : float
            geocent_time
        params[9] : float
            a_1
        params[10] : float
            a_2
        params[11] : float
            tilt_1
        params[12] : float
            tilt_2
        params[13] : float
            phi_12
        params[14] : float
            phi_jl
        params[15] : float
            lambda_1
        params[16] : float
            lambda_2
        params[17] : float
            eccentricity
        params[18] : float
            approximant
        params[19] : float
            f_min
        params[20] : float
            f_ref
        params[21] : float
            duration
        params[22] : float
            sampling_frequency
        params[23] : int
            index tracker
        params[24] : list
            list of psds for each detector
        params[25] : str
            frequency_domain_source_model name
        params[26] : list or None
            noise realization. If None, then PSD as noise realization
        

    Returns
    -------
    SNRs_list : list
        contains opt_snr for each detector and net_opt_snr
    params[22] : int
        index tracker
    """

    bilby.core.utils.logger.disabled = True
    np.random.seed(88170235)
    parameters = {
        "mass_1": params[0],
        "mass_2": params[1],
        "luminosity_distance": params[2],
        "theta_jn": params[3],
        "psi": params[4],
        "phase": params[5],
        "geocent_time": params[8],
        "ra": params[6],
        "dec": params[7],
        "a_1": params[9],
        "a_2": params[10],
        "tilt_1": params[11],
        "tilt_2": params[12],
        "phi_12": params[13],
        "phi_jl": params[14],
        "lambda_1": params[15],
        "lambda_2": params[16],
        "eccentricity": params[17],
    }

    # print('eccentricity', params[17])
    # print('frequency_domain_source_model', params[25])

    waveform_arguments = dict(
        waveform_approximant=params[18],
        reference_frequency=params[20],
        minimum_frequency=params[19],
    )

    waveform_generator = bilby.gw.WaveformGenerator(
        duration=params[21],
        sampling_frequency=params[22],
        frequency_domain_source_model=getattr(bilby.gw.source, params[25]),
        waveform_arguments=waveform_arguments,
    )
    polas = waveform_generator.frequency_domain_strain(parameters=parameters)

    # h = F+.h+ + Fx.hx
    # <h|h> = F+^2<h+,h+> + Fx^2<hx,hx> + F+Fx 2<h+,hx>
    # <h|h> = F+^2<h+,h+> + Fx^2<hx,hx>, if h+ and hx are orthogonal
    # F+^2 and Fx^2 will be added later
    hp_inner_hp_list = []
    hc_inner_hc_list = []
    n_inner_hp_list = []
    n_inner_hc_list = []
    # list_of_detectors = params[26:].tolist()
    psds_objects = params[24]
    noise = params[26]
    for idx in range(len(psds_objects)):
    # for idx, det in enumerate(list_of_detectors):

        # need to compute the inner product for
        p_array = psds_objects[idx][2](waveform_generator.frequency_array)
        if noise is None:
            noise = np.sqrt(p_array)
        else:
            noise = np.array(noise)

        idx2 = (noise != 0.0) & (noise != np.inf)
        hp_inner_hp = noise_weighted_inner_product(
            polas["plus"][idx2],
            polas["plus"][idx2],
            p_array[idx2],
            waveform_generator.duration,
        )
        hc_inner_hc = noise_weighted_inner_product(
            polas["cross"][idx2],
            polas["cross"][idx2],
            p_array[idx2],
            waveform_generator.duration,
        )
        n_inner_hp = noise_weighted_inner_product(
            noise[idx2],
            polas["plus"][idx2],
            p_array[idx2],
            waveform_generator.duration,
        )
        n_inner_hc = noise_weighted_inner_product(
            noise[idx2],
            polas["cross"][idx2],
            p_array[idx2],
            waveform_generator.duration,
        )
        # if idx == 2:
        #     # save n_inner_hp and n_inner_hc
        #     np.save("n_inner_hp.npy", n_inner_hp)
        #     np.save("n_inner_hc.npy", n_inner_hc)
        #     np.save("noise.npy", noise[idx2])
        #     np.save("p_array.npy", p_array[idx2])
        #     np.save("polas_plus.npy", polas["plus"][idx2])
        #     np.save("polas_cross.npy", polas["cross"][idx2])
        #     np.save("duration.npy", waveform_generator.duration)

        hp_inner_hp_list.append(hp_inner_hp)
        hc_inner_hc_list.append(hc_inner_hc)
        n_inner_hp_list.append(n_inner_hp)
        n_inner_hc_list.append(n_inner_hc)

    return (hp_inner_hp_list, hc_inner_hc_list, n_inner_hp_list, n_inner_hc_list, params[23])

def noise_weighted_inner_prod_ripple(params):
    """
        Probaility of detection of GW for the given sensitivity of the detectors

        Parameters
        ----------
        params : list
            list of parameters for the inner product calculation
            List contains: \n
            params[0] : `numpy.ndarray`
                plus polarization
            params[1] : `numpy.ndarray`
                cross polarization
            params[2] : `numpy.ndarray`
                frequency array
            params[3] : `float`
                cutt-off size of given arrays
            params[4] : `float`
                minimum frequency
            params[5] : `float`
                duration
            params[6] : `int`
                index
            params[7] : `list` 
                psd objects of given detectors

        Returns
        -------
        SNRs_list : list
            contains opt_snr for each detector and net_opt_snr
        params[22] : int
            index tracker
    """

    ## input 
    hp = params[0]
    hc = params[1]
    fs = params[2]
    fsize = params[3]
    fmin = params[4]
    duration = params[5]

    # for i in range(size):
    # remove the np.nan padding
    hp = np.array(hp[:fsize])
    hc = np.array(hc[:fsize])
    # find the index of 20Hz or nearby
    # set all elements to zero below this index
    fs =  fs[:fsize]
    idx = np.abs( fs - fmin).argmin()
    hp[:idx] = 0.0 + 0.0j
    hc[:idx] = 0.0 + 0.0j

    # h = F+.h+ + Fx.hx
    # <h|h> = <h+,h+> + <hx,hx> + 2<h+,hx>
    # <h|h> = <h+,h+> + <hx,hx>, if h+ and hx are orthogonal
    hp_inner_hp_list = []
    hc_inner_hc_list = []
    psds_objects = params[7]
    
    for idx in range(len(psds_objects)):

        # need to compute the inner product for
        p_array = psds_objects[idx][2](fs)

        idx2 = (p_array != 0.0) & (p_array != np.inf)
        # complex128 is not available in JAX numpy operation.
        # That's why I will use numba njit function
        hp_inner_hp = noise_weighted_inner_product(
            hp[idx2],
            hp[idx2],
            p_array[idx2],
            duration,
        )
        # complex128 is not available in JAX numpy operation.
        # That's why I will use numba njit function
        hc_inner_hc = noise_weighted_inner_product(
            hc[idx2],
            hc[idx2],
            p_array[idx2],
            duration,
        )

        hp_inner_hp_list.append(hp_inner_hp)
        hc_inner_hc_list.append(hc_inner_hc)
    
    return (hp_inner_hp_list, hc_inner_hc_list, params[6])





from .njit_functions import (
    findchirp_chirptime,
    antenna_response_plus,
    antenna_response_cross,
    antenna_response_array,
    noise_weighted_inner_product,
    effective_distance,
    effective_distance_array,
    linear_interpolator
)

from .njit_interpolators import (
    get_interpolated_snr_aligned_spins_numba,
    get_interpolated_snr_no_spins_numba,
)

# __all__ = [
#     # njit_functions
#     'findchirp_chirptime',
#     'antenna_response_plus',
#     'antenna_response_cross',
#     'antenna_response_array',
#     'noise_weighted_inner_product',
#     'effective_distance',
#     'effective_distance_array',
#     'linear_interpolator',
#     # njit_interpolators
#     'get_interpolated_snr_aligned_spins_numba',
#     'get_interpolated_snr_no_spins_numba',
# ]


import mlx.core as mx
import numpy as np
from ..numba import antenna_response_array

@mx.compile
def find_index_1d_mlx(x_array, x_new):
    N = x_array.shape[0]
    i = mx.sum(x_array <= x_new) - 1
    i = mx.clip(i, 1, N - 3)
    condition_i = mx.where(x_new <= x_array[1], 1,
                  mx.where(x_new >= x_array[N-2], 3, 2))
    return i, condition_i

@mx.compile
def spline_interp_4pts_mlx(x_eval, x_pts, y_pts, condition_i):
    
    # Linear at boundaries
    def left_branch(x_eval, x_pts, y_pts):
        denom_left = x_pts[1] - x_pts[0]
        left_val = y_pts[0] + (y_pts[1] - y_pts[0]) * (x_eval - x_pts[0]) / denom_left
        return left_val

    def right_branch(x_eval, x_pts, y_pts):
        denom_right = x_pts[3] - x_pts[2]
        right_val = y_pts[2] + (y_pts[3] - y_pts[2]) * (x_eval - x_pts[2]) / denom_right
        return right_val

    # Cubic in middle
    def cubic_branch(x_eval, x_pts, y_pts):
        x0, x1, x2, x3 = x_pts[0], x_pts[1], x_pts[2], x_pts[3]
        y0, y1, y2, y3 = y_pts[0], y_pts[1], y_pts[2], y_pts[3]
        denom = x2 - x1
        t = (x_eval - x1) / denom

        m1 = ((y2 - y1) / (x2 - x1)) * ((x1 - x0) / (x2 - x0)) + ((y1 - y0) / (x1 - x0)) * ((x2 - x1) / (x2 - x0))
        m2 = ((y3 - y2) / (x3 - x2)) * ((x2 - x1) / (x3 - x1)) + ((y2 - y1) / (x2 - x1)) * ((x3 - x2) / (x3 - x1))

        h00 = (2 * t**3 - 3 * t**2 + 1)
        h10 = (t**3 - 2 * t**2 + t)
        h01 = (-2 * t**3 + 3 * t**2)
        h11 = (t**3 - t**2)

        cubic_val = h00 * y1 + h10 * m1 * denom + h01 * y2 + h11 * m2 * denom
        return cubic_val

    # Use chained mx.where for multi-way select
    return mx.where(condition_i == 1, left_branch(x_eval, x_pts, y_pts),
                   mx.where(condition_i == 3, right_branch(x_eval, x_pts, y_pts), cubic_branch(x_eval, x_pts, y_pts)))

######### Interpolation 4D #########
@mx.compile
def spline_interp_4x4x4x4pts_mlx(q_array, mtot_array, a1_array, a2_array, snrpartialscaled_array, q_new, mtot_new, a1_new, a2_new):
    """
    Helper function that performs the FULL 4D interpolation for a SINGLE point.
    This function finds indices, slices data, and then uses vmap internally
    to perform interpolation efficiently without Python loops.
    """
    # 1. Find indices and conditions for the new point
    q_idx, int_q = find_index_1d_mlx(q_array, q_new)
    m_idx, int_m = find_index_1d_mlx(mtot_array, mtot_new)
    a1_idx, int_a1 = find_index_1d_mlx(a1_array, a1_new)
    a2_idx, int_a2 = find_index_1d_mlx(a2_array, a2_new)

    # 2. Slice the small 4x4x4x4 cube of data and the corresponding 4-point coordinate arrays
    q_pts = mx.array([q_array[q_idx - 1], q_array[q_idx], q_array[q_idx + 1], q_array[q_idx + 2]])
    m_pts = mx.array([mtot_array[m_idx - 1], mtot_array[m_idx], mtot_array[m_idx + 1], mtot_array[m_idx + 2]])
    a1_pts = mx.array([a1_array[a1_idx - 1], a1_array[a1_idx], a1_array[a1_idx + 1], a1_array[a1_idx + 2]])
    a2_pts = mx.array([a2_array[a2_idx - 1], a2_array[a2_idx], a2_array[a2_idx + 1], a2_array[a2_idx + 2]])

    # # --- Create the 4x4x4x4 data_cube ---
    # # We expand the index arrays with new dimensions so they broadcast correctly
    # # to select a 4x4x4x4 cube from the larger array.
    offsets = mx.arange(-1, 3)
    q_indices  = (q_idx + offsets).reshape(4, 1, 1, 1)
    m_indices  = (m_idx + offsets).reshape(1, 4, 1, 1)
    a1_indices = (a1_idx + offsets).reshape(1, 1, 4, 1)
    a2_indices = (a2_idx + offsets).reshape(1, 1, 1, 4)

    # Perform a single, efficient "gather" operation using the broadcasted indices.
    data_cube = snrpartialscaled_array[q_indices, m_indices, a1_indices, a2_indices]


    # 3. Perform 4D interpolation using vmap to eliminate loops
    # Interpolate along the last dimension (a2) for each of the 4x4x4 slices
    interp_on_a2 = mx.vmap(
        lambda y_slice: spline_interp_4pts_mlx(a2_new, a2_pts, y_slice, int_a2)
    )(data_cube.reshape(64, 4)).reshape(4, 4, 4)

    # Interpolate along the next dimension (a1)
    interp_on_a1 = mx.vmap(
        lambda y_slice: spline_interp_4pts_mlx(a1_new, a1_pts, y_slice, int_a1)
    )(interp_on_a2.reshape(16, 4)).reshape(4, 4)

    # Interpolate along the mtot dimension
    interp_on_m = mx.vmap(
        lambda y_slice: spline_interp_4pts_mlx(mtot_new, m_pts, y_slice, int_m)
    )(interp_on_a1)

    # Final interpolation along the q dimension
    final_snr = spline_interp_4pts_mlx(q_new, q_pts, interp_on_m, int_q)

    return final_snr

@mx.compile
def spline_interp_4x4x4x4pts_batched_mlx(q_array, mtot_array, a1_array, a2_array, snrpartialscaled_array, q_new_batch, mtot_new_batch, a1_new_batch, a2_new_batch):
    """
    Perform batched 4D cubic spline interpolation using JAX vectorization.
    """
    # Vectorize the complete single-point interpolation function.
    # This is the only vmap call needed at the top level.
    vmapped_interpolator = mx.vmap(
        spline_interp_4x4x4x4pts_mlx,
        in_axes=(None, None, None, None, None, 0, 0, 0, 0)
    )

    # Call the fully vectorized function on the batch data.
    return vmapped_interpolator(
        q_array, mtot_array, a1_array, a2_array, snrpartialscaled_array,
        q_new_batch, mtot_new_batch, a1_new_batch, a2_new_batch
    )

def get_interpolated_snr_aligned_spins_mlx(mass_1, mass_2, luminosity_distance, theta_jn, psi, geocent_time, ra, dec, a_1, a_2, detector_tensor, snr_partialscaled, ratio_arr, mtot_arr, a1_arr, a_2_arr, batch_size=100000):
    """
    Calculate interpolated signal-to-noise ratio (SNR) for aligned spin gravitational wave signals using JAX.
    This function computes the SNR for gravitational wave signals with aligned spins across multiple 
    detectors using 4D cubic spline interpolation. It calculates the effective distance, partial SNR,
    and combines results from multiple detectors to produce the effective SNR.

    Parameters
    ----------
    mass_1 : jax.numpy.ndarray
        Primary mass of the binary system in solar masses.
    mass_2 : jax.numpy.ndarray  
        Secondary mass of the binary system in solar masses.
    luminosity_distance : jax.numpy.ndarray
        Luminosity distance to the source in Mpc.
    theta_jn : jax.numpy.ndarray
        Inclination angle between the orbital angular momentum and line of sight in radians.
    psi : jax.numpy.ndarray
        Polarization angle in radians.
    geocent_time : jax.numpy.ndarray
        GPS time of coalescence at the geocenter in seconds.
    ra : jax.numpy.ndarray
        Right ascension of the source in radians.
    dec : jax.numpy.ndarray
        Declination of the source in radians.
    a_1 : jax.numpy.ndarray
        Dimensionless spin magnitude of the primary black hole.
    a_2 : jax.numpy.ndarray
        Dimensionless spin magnitude of the secondary black hole.
    detector_tensor : jax.numpy.ndarray
        Detector tensor array containing detector response information.
        Shape: (n_detectors, ...)
    snr_partialscaled : jax.numpy.ndarray
        Pre-computed scaled partial SNR values for interpolation.
        Shape: (n_detectors, ...)
    ratio_arr : jax.numpy.ndarray
        Mass ratio grid points for interpolation (q = m2/m1).
    mtot_arr : jax.numpy.ndarray
        Total mass grid points for interpolation.
    a1_arr : jax.numpy.ndarray
        Primary spin grid points for interpolation.
    a_2_arr : jax.numpy.ndarray
        Secondary spin grid points for interpolation.

    Returns
    -------
    snr : jax.numpy.ndarray
        SNR values for each detector. Shape: (n_detectors, n_samples)
    snr_effective : jax.numpy.ndarray
        Effective SNR combining all detectors. Shape: (n_samples,)
    snr_partial_ : jax.numpy.ndarray
        Interpolated partial SNR values for each detector. Shape: (n_detectors, n_samples)
    d_eff : jax.numpy.ndarray
        Effective distance for each detector accounting for antenna response.
        Shape: (n_detectors, n_samples)

    Notes
    -----
    - Uses 4D cubic spline interpolation for efficient SNR calculation
    - Assumes aligned spins (no precession)
    - Effective SNR is calculated as sqrt(sum(SNR_i^2)) across detectors
    - Chirp mass and inclination-dependent factors are computed analytically
    """

    # get array of antenna response
    # antenna_response_array is already compiled with numba
    # this is done so as to preserve the float64 precesion in the antenna response calculation
    Fp, Fc = antenna_response_array(ra, dec, geocent_time, psi, detector_tensor)

    snr, snr_effective, snr_partial_, d_eff = get_interpolated_snr_aligned_spins_helper(
        mx.array(mass_1), 
        mx.array(mass_2), 
        mx.array(luminosity_distance), 
        mx.array(theta_jn), 
        mx.array(a_1), 
        mx.array(a_2), 
        mx.array(snr_partialscaled), 
        mx.array(ratio_arr), 
        mx.array(mtot_arr), 
        mx.array(a1_arr), 
        mx.array(a_2_arr),
        mx.array(Fp),
        mx.array(Fc),
        mx.array(detector_tensor),
        batch_size
    )

    return np.array(snr), np.array(snr_effective), np.array(snr_partial_), np.array(d_eff)

@mx.compile
def get_interpolated_snr_aligned_spins_helper(mass_1, mass_2, luminosity_distance, theta_jn, a_1, a_2, snr_partialscaled, ratio_arr, mtot_arr, a1_arr, a_2_arr, Fp, Fc, detector_tensor, batch_size):
    """
    Calculate interpolated signal-to-noise ratio (SNR) for aligned spin gravitational wave signals using JAX.
    This function computes the SNR for gravitational wave signals with aligned spins across multiple 
    detectors using 4D cubic spline interpolation. It calculates the effective distance, partial SNR,
    and combines results from multiple detectors to produce the effective SNR.

    Parameters
    ----------
    mass_1 : jax.numpy.ndarray
        Primary mass of the binary system in solar masses.
    mass_2 : jax.numpy.ndarray  
        Secondary mass of the binary system in solar masses.
    luminosity_distance : jax.numpy.ndarray
        Luminosity distance to the source in Mpc.
    theta_jn : jax.numpy.ndarray
        Inclination angle between the orbital angular momentum and line of sight in radians.
    psi : jax.numpy.ndarray
        Polarization angle in radians.
    geocent_time : jax.numpy.ndarray
        GPS time of coalescence at the geocenter in seconds.
    ra : jax.numpy.ndarray
        Right ascension of the source in radians.
    dec : jax.numpy.ndarray
        Declination of the source in radians.
    a_1 : jax.numpy.ndarray
        Dimensionless spin magnitude of the primary black hole.
    a_2 : jax.numpy.ndarray
        Dimensionless spin magnitude of the secondary black hole.
    detector_tensor : jax.numpy.ndarray
        Detector tensor array containing detector response information.
        Shape: (n_detectors, ...)
    snr_partialscaled : jax.numpy.ndarray
        Pre-computed scaled partial SNR values for interpolation.
        Shape: (n_detectors, ...)
    ratio_arr : jax.numpy.ndarray
        Mass ratio grid points for interpolation (q = m2/m1).
    mtot_arr : jax.numpy.ndarray
        Total mass grid points for interpolation.
    a1_arr : jax.numpy.ndarray
        Primary spin grid points for interpolation.
    a_2_arr : jax.numpy.ndarray
        Secondary spin grid points for interpolation.

    Returns
    -------
    snr : jax.numpy.ndarray
        SNR values for each detector. Shape: (n_detectors, n_samples)
    snr_effective : jax.numpy.ndarray
        Effective SNR combining all detectors. Shape: (n_samples,)
    snr_partial_ : jax.numpy.ndarray
        Interpolated partial SNR values for each detector. Shape: (n_detectors, n_samples)
    d_eff : jax.numpy.ndarray
        Effective distance for each detector accounting for antenna response.
        Shape: (n_detectors, n_samples)

    Notes
    -----
    - Uses 4D cubic spline interpolation for efficient SNR calculation
    - Assumes aligned spins (no precession)
    - Effective SNR is calculated as sqrt(sum(SNR_i^2)) across detectors
    - Chirp mass and inclination-dependent factors are computed analytically
    """
    
    size = mass_1.shape[0]
    det_len = detector_tensor.shape[0]
    
    mtot = mass_1 + mass_2
    ratio = mass_2 / mass_1

    Mc = ((mass_1 * mass_2) ** (3 / 5)) / ((mass_1 + mass_2) ** (1 / 5))
    A1 = Mc ** (5.0 / 6.0)
    ci_2 = mx.cos(theta_jn) ** 2
    ci_param = ((1 + mx.cos(theta_jn) ** 2) / 2) ** 2

    snr_partial_ = mx.zeros((det_len,size))
    d_eff = mx.zeros((det_len,size))
    snr = mx.zeros((det_len,size))
    # loop over the detectors
    for j in range(det_len):
        # Iterate over the data in batches
        for i in range(0, size, batch_size):
            # Define the start and end indices for the current batch
            start_idx = i
            end_idx = min(i + batch_size, size)

            snr_partial_[j, start_idx:end_idx] = mx.array(spline_interp_4x4x4x4pts_batched_mlx(
                q_array=ratio_arr,
                mtot_array=mtot_arr,
                a1_array=a1_arr,
                a2_array=a_2_arr,
                snrpartialscaled_array=snr_partialscaled[j],
                q_new_batch=ratio,
                mtot_new_batch=mtot,
                a1_new_batch=a_1,
                a2_new_batch=a_2,
            )
        )

        d_eff[j] = luminosity_distance / mx.sqrt(Fp[j] ** 2 * ci_param + Fc[j] ** 2 * ci_2)
        # Calculate the SNR for this detector
        snr[j] = snr_partial_[j] * A1 / d_eff[j]

    snr_effective = mx.sqrt(mx.sum(snr ** 2, axis=0))

    return snr, snr_effective, snr_partial_, d_eff


######### Interpolation 2D #########
@mx.compile
def spline_interp_4x4pts_mlx(q_array, mtot_array, snrpartialscaled_array, q_new, mtot_new):
    """
    Helper function that performs the FULL 2D interpolation for a SINGLE point.
    This function finds indices, slices data, and then uses vmap internally
    to perform interpolation efficiently without Python loops.
    """
    # 1. Find indices and conditions for the new point
    q_idx, int_q = find_index_1d_mlx(q_array, q_new)
    m_idx, int_m = find_index_1d_mlx(mtot_array, mtot_new)

    # 2. Slice the small 4x4 cube of data and the corresponding 4-point coordinate arrays
    q_pts = mx.array([q_array[q_idx - 1], q_array[q_idx], q_array[q_idx + 1], q_array[q_idx + 2]])
    m_pts = mx.array([mtot_array[m_idx - 1], mtot_array[m_idx], mtot_array[m_idx + 1], mtot_array[m_idx + 2]])

    # # --- Create the 4x4 data_cube ---
    # # We expand the index arrays with new dimensions so they broadcast correctly
    # # to select a 4x4 cube from the larger array.
    offsets = mx.arange(-1, 3)
    q_indices  = (q_idx + offsets).reshape(4, 1)
    m_indices  = (m_idx + offsets).reshape(1, 4)

    # Perform a single, efficient "gather" operation using the broadcasted indices.
    data_cube = snrpartialscaled_array[q_indices, m_indices]

    # 3. Perform 2D interpolation using vmap to eliminate loops
    # Interpolate along the mtot dimension
    interp_on_m = mx.vmap(
        lambda y_slice: spline_interp_4pts_mlx(mtot_new, m_pts, y_slice, int_m)
    )(data_cube)

    # Final interpolation along the q dimension
    final_snr = spline_interp_4pts_mlx(q_new, q_pts, interp_on_m, int_q)

    return final_snr

@mx.compile
def spline_interp_4x4pts_batched_mlx(q_array, mtot_array, snrpartialscaled_array, q_new_batch, mtot_new_batch):
    """
    Perform batched 2D cubic spline interpolation using MLX vectorization.
    """
    # Vectorize the complete single-point interpolation function.
    # This is the only vmap call needed at the top level.
    vmapped_interpolator = mx.vmap(
        spline_interp_4x4pts_mlx,
        in_axes=(None, None, None, 0, 0)
    )

    # Call the fully vectorized function on the batch data.
    return vmapped_interpolator(
        q_array, mtot_array, snrpartialscaled_array,
        q_new_batch, mtot_new_batch
    )

def get_interpolated_snr_no_spins_mlx(mass_1, mass_2, luminosity_distance, theta_jn, psi, geocent_time, ra, dec, a_1, a_2, detector_tensor, snr_partialscaled, ratio_arr, mtot_arr, a1_arr, a_2_arr, batch_size=100000):
    """
    Calculate interpolated signal-to-noise ratio (SNR) for aligned spin gravitational wave signals using JAX.
    This function computes the SNR for gravitational wave signals with aligned spins across multiple 
    detectors using 4D cubic spline interpolation. It calculates the effective distance, partial SNR,
    and combines results from multiple detectors to produce the effective SNR.

    Parameters
    ----------
    mass_1 : jax.numpy.ndarray
        Primary mass of the binary system in solar masses.
    mass_2 : jax.numpy.ndarray  
        Secondary mass of the binary system in solar masses.
    luminosity_distance : jax.numpy.ndarray
        Luminosity distance to the source in Mpc.
    theta_jn : jax.numpy.ndarray
        Inclination angle between the orbital angular momentum and line of sight in radians.
    psi : jax.numpy.ndarray
        Polarization angle in radians.
    geocent_time : jax.numpy.ndarray
        GPS time of coalescence at the geocenter in seconds.
    ra : jax.numpy.ndarray
        Right ascension of the source in radians.
    dec : jax.numpy.ndarray
        Declination of the source in radians.
    a_1 : jax.numpy.ndarray
        Dimensionless spin magnitude of the primary black hole.
    a_2 : jax.numpy.ndarray
        Dimensionless spin magnitude of the secondary black hole.
    detector_tensor : jax.numpy.ndarray
        Detector tensor array containing detector response information.
        Shape: (n_detectors, ...)
    snr_partialscaled : jax.numpy.ndarray
        Pre-computed scaled partial SNR values for interpolation.
        Shape: (n_detectors, ...)
    ratio_arr : jax.numpy.ndarray
        Mass ratio grid points for interpolation (q = m2/m1).
    mtot_arr : jax.numpy.ndarray
        Total mass grid points for interpolation.
    a1_arr : jax.numpy.ndarray
        Primary spin grid points for interpolation.
    a_2_arr : jax.numpy.ndarray
        Secondary spin grid points for interpolation.

    Returns
    -------
    snr : jax.numpy.ndarray
        SNR values for each detector. Shape: (n_detectors, n_samples)
    snr_effective : jax.numpy.ndarray
        Effective SNR combining all detectors. Shape: (n_samples,)
    snr_partial_ : jax.numpy.ndarray
        Interpolated partial SNR values for each detector. Shape: (n_detectors, n_samples)
    d_eff : jax.numpy.ndarray
        Effective distance for each detector accounting for antenna response.
        Shape: (n_detectors, n_samples)

    Notes
    -----
    - Uses 4D cubic spline interpolation for efficient SNR calculation
    - Assumes aligned spins (no precession)
    - Effective SNR is calculated as sqrt(sum(SNR_i^2)) across detectors
    - Chirp mass and inclination-dependent factors are computed analytically
    """

    # get array of antenna response
    # antenna_response_array is already compiled with numba
    # this is done so as to preserve the float64 precesion in the antenna response calculation
    Fp, Fc = antenna_response_array(ra, dec, geocent_time, psi, detector_tensor)

    snr, snr_effective, snr_partial_, d_eff = get_interpolated_snr_no_spins_helper(
        mx.array(mass_1), 
        mx.array(mass_2), 
        mx.array(luminosity_distance), 
        mx.array(theta_jn), 
        mx.array(snr_partialscaled), 
        mx.array(ratio_arr), 
        mx.array(mtot_arr), 
        mx.array(Fp),
        mx.array(Fc),
        mx.array(detector_tensor),
        batch_size
    )

    return np.array(snr), np.array(snr_effective), np.array(snr_partial_), np.array(d_eff)

@mx.compile
def get_interpolated_snr_no_spins_helper(mass_1, mass_2, luminosity_distance, theta_jn, snr_partialscaled, ratio_arr, mtot_arr, Fp, Fc, detector_tensor, batch_size):
    """
    Calculate interpolated signal-to-noise ratio (SNR) for non-spinning gravitational wave signals using JAX.
    This function computes the SNR for gravitational wave signals without spins across multiple 
    detectors using 4D cubic spline interpolation. It calculates the effective distance, partial SNR,
    and combines results from multiple detectors to produce the effective SNR.
    """
    
    size = mass_1.shape[0]
    det_len = detector_tensor.shape[0]
    
    mtot = mass_1 + mass_2
    ratio = mass_2 / mass_1

    Mc = ((mass_1 * mass_2) ** (3 / 5)) / ((mass_1 + mass_2) ** (1 / 5))
    A1 = Mc ** (5.0 / 6.0)
    ci_2 = mx.cos(theta_jn) ** 2
    ci_param = ((1 + mx.cos(theta_jn) ** 2) / 2) ** 2

    snr_partial_ = mx.zeros((det_len,size))
    d_eff = mx.zeros((det_len,size))
    snr = mx.zeros((det_len,size))
    # loop over the detectors
    for j in range(det_len):
        # Iterate over the data in batches
        for i in range(0, size, batch_size):
            # Define the start and end indices for the current batch
            start_idx = i
            end_idx = min(i + batch_size, size)
            
            snr_partial_[j, start_idx:end_idx] = mx.array(spline_interp_4x4pts_batched_mlx(
                    q_array=ratio_arr,
                    mtot_array=mtot_arr,
                    snrpartialscaled_array=snr_partialscaled[j],
                    q_new_batch=ratio,
                    mtot_new_batch=mtot
                )
            )

        d_eff[j] = luminosity_distance / mx.sqrt(Fp[j] ** 2 * ci_param + Fc[j] ** 2 * ci_2)
        # Calculate the SNR for this detector
        snr[j] = snr_partial_[j] * A1 / d_eff[j]

    snr_effective = mx.sqrt(mx.sum(snr ** 2, axis=0))

    return snr, snr_effective, snr_partial_, d_eff


"""
This module implements the SNRThresholdFinder class to determine the optimal SNR threshold for gravitational wave detection using cross-entropy maximization (following Essick et al. 2023).
"""
import multiprocessing as mpp
from tqdm import tqdm
import h5py
import numpy as np
from .crossentropydifference import cross_entropy_difference
# import crossentropydifference
# cross_entropy_difference = crossentropydifference.cross_entropy_difference
from scipy.interpolate import interp1d
from scipy.optimize import minimize_scalar


class SNRThresholdFinder:
    """
    A class to find the optimal SNR threshold for gravitational wave detection using cross-entropy maximization.

    Parameters
    ----------
    catalog_file : str
        Path to the HDF5 file containing the injection catalog data. The file should have something like the following structure (refer to https://zenodo.org/records/16740117):
        ```
        injections.hdf
        |-- events
        |   |-- z  (parameter to me fitted on)
        |   |-- mass1_source (parameter with which the data is to be selected with)
        |   |-- gstlal_far (original_detection_statistic)
        |   |-- observed_snr_net (projected_detection_statistic)
        ```
    original_detection_statistic : dict, optional
        Dictionary specifying the original detection statistic with keys:
        'parameter' (str): Name of the key in the catalog for the original detection statistic.
        'threshold' (float): Threshold value for the original detection statistic.
        Default is {'parameter': 'gstlal_far', 'threshold': 1}.
    projected_detection_statistic : dict, optional
        Dictionary specifying the projected detection statistic with keys: 
        'parameter' (str): Name of the key in the catalog for the projected detection statistic.
        'threshold' (float): Threshold value for the projected detection statistic.
        'threshold_search_bounds' (tuple): Bounds for the threshold search.
        Default is {'parameter': 'observed_snr_net', 'threshold': None, 'threshold_search_bounds': (4, 14)}.
    parameters_to_fit : list of str, optional
        List of parameter to fit, e.g., ['redshift']. Default is ['redshift'].
    sample_size : int, optional
        Number of samples to use for KDE estimation. Default is 10000.
    selection_range : dict, optional
        Dictionary specifying the selection range with keys:
        'parameter' (str or list): Parameter(s) to apply the selection range on.
        'range' (tuple): Tuple specifying the (min, max) range for selection.
        Default is {'parameter': 'mass1_source', 'range': (5, 200)}.

    Examples
    ----------
    >>> finder = SNRThresholdFinder(catalog_file='injection_catalog.h5')
    >>> best_thr, del_H, H, H_true, snr_thrs = finder.find_threshold(iteration=10)
    >>> print(f"Best SNR threshold: {best_thr:.2f}")

    Instance Attributes
    ----------
    SNRThresholdFinder class has the following attributes, \n
    +-------------------------------------+----------------------------------+
    | Atrributes                          | Type                             |
    +=====================================+==================================+
    | original_detection_statistic       | dict                             |
    +-------------------------------------+----------------------------------+
    | projected_detection_statistic      | dict                             |
    +-------------------------------------+----------------------------------+
    | parameters_to_fit                 | list                             |
    +-------------------------------------+----------------------------------+
    | sample_size                        | int                              |
    +-------------------------------------+----------------------------------+
    | selection_range                    | dict                             |
    +-------------------------------------+----------------------------------+

    Instance Methods
    ----------
    GWSNR class has the following methods, \n
    +-------------------------------------+----------------------------------+
    | Methods                             | Description                      |
    +=====================================+==================================+
    | det_data                            | Load and preprocess catalog data |
    +-------------------------------------+----------------------------------+
    | find_threshold                      | Find the optimal SNR threshold   |
    +-------------------------------------+----------------------------------+
    | find_best_SNR_threshold             | Find the best SNR threshold using |
    |                                     | spline interpolation and         |
    |                                     | optimization                     |
    +-------------------------------------+----------------------------------+
    """

    def __init__(self, 
                catalog_file=None, 
                npool=4,
                multiprocessing_verbose=True,
                original_detection_statistic=None, projected_detection_statistic=None,
                parameters_to_fit=None,
                sample_size=20000,
                selection_range=None,
        ):

        self.npool = npool
        self.multiprocessing_verbose = multiprocessing_verbose

        if selection_range is None:
            selection_range = dict(
                key_name = 'mass1_source',
                parameter = None,
                range = (30, 60),
            )
        self.selection_range = selection_range

        if original_detection_statistic is None:
            self.original_detection_statistic = dict(
                key_name='gstlal_far',
                parameter=None,
                threshold=1,  # 1 per year
            )
        else:
            self.original_detection_statistic = original_detection_statistic

        if projected_detection_statistic is None:
            self.projected_detection_statistic = dict(
                key_name='observed_snr_net',
                parameter=None,
                threshold=None, # to be determined
                threshold_search_bounds=(4, 14),
            )
        else:
            self.projected_detection_statistic = projected_detection_statistic

        if parameters_to_fit is None:
            self.parameters_to_fit = dict(
                key_name = 'z',
                parameter = None,
            )
        else:
            self.parameters_to_fit = parameters_to_fit

        self.sample_size = sample_size

        self.det_data(catalog_file);

    def det_data(self, 
            catalog_file,
        ):
        """
        Function to load and preprocess the injection catalog data from an HDF5 file.

        Parameters
        ----------
        catalog_file : str
            Path to the HDF5 file containing the injection catalog data.
        Returns
        -------
        result_dict : dict
            Dictionary containing the preprocessed data for the specified parameters and detection statistics.

        Raises
        ------
        ValueError
            If 'redshift' is not included in parameters_to_fit.
        """
        def raise_not_provided(param):
            raise ValueError(f"if catalog_file is not provided, you must provide {param} as list or numpy array.")
        
        if catalog_file is None:
            param = self.selection_range['parameter']
            if isinstance(param, list) or isinstance(param, np.ndarray):
                self.selection_range['parameter'] = np.array(param)
            else:
                raise_not_provided('selection_range["parameter"]')
            
            param = self.original_detection_statistic['parameter']
            if isinstance(param, list) or isinstance(param, np.ndarray):
                self.original_detection_statistic['parameter'] = np.array(param)
            else:
                raise_not_provided('original_detection_statistic["parameter"]')

            param = self.projected_detection_statistic['parameter']
            if isinstance(param, list) or isinstance(param, np.ndarray):
                self.projected_detection_statistic['parameter'] = np.array(param)
            else:
                raise_not_provided('projected_detection_statistic["parameter"]')

            param = self.parameters_to_fit['parameter']
            if isinstance(param, list) or isinstance(param, np.ndarray):
                self.parameters_to_fit['parameter'] = np.array(param)
            else:
                raise_not_provided('parameters_to_fit["parameter"]')

        else:
            with h5py.File(catalog_file, 'r') as obj:
                attrs = dict(obj.attrs.items())
                events = obj['events'][:]

            key_name = self.selection_range['key_name']
            if key_name in events.dtype.names:
                self.selection_range['parameter'] = events[key_name]
            else:
                print(f"[WARNING] {key_name} not found in the catalog. Using the parameter array of the same name if provided.")

            key_name = self.original_detection_statistic['key_name']
            self.original_detection_statistic['parameter'] = events[key_name]

            key_name = self.projected_detection_statistic['key_name']
            self.projected_detection_statistic['parameter'] = events[key_name]

            key_name = self.parameters_to_fit['key_name']
            if isinstance(key_name, list):
                param_array = []
                for i, kn in enumerate(key_name):
                    if isinstance(self.parameters_to_fit['parameter'], list) or isinstance(self.parameters_to_fit['parameter'], np.ndarray):
                        param_array.append(self.parameters_to_fit['parameter'][i])
                    elif kn in events.dtype.names:
                        param_array.append(events[kn])
                    else:
                        raise ValueError(f"{kn} not found in the catalog. Please provide the parameter array of the same name.")
                self.parameters_to_fit['parameter'] = np.array(param_array)
            else:
                self.parameters_to_fit['parameter'] = events[key_name]

        # select only events within the selection range
        min_val = self.selection_range['range'][0]
        max_val = self.selection_range['range'][1]
        param = self.selection_range['parameter']
        idx_ = (param >= min_val) & (param <= max_val)

        dim = len(self.parameters_to_fit['parameter'].shape)
        if dim < 2:
            self.parameters_to_fit['parameter'] = self.parameters_to_fit['parameter'][idx_]
        else:
            raise NotImplementedError("Selection range filtering for multi-dimensional parameters_to_fit is not implemented yet.")
            # param_array = []
            # for i in range(dim):
            #     param_array.append(self.parameters_to_fit['parameter'][i][idx_])
            # self.parameters_to_fit['parameter'] = np.array(param_array)
        self.original_detection_statistic['parameter'] = self.original_detection_statistic['parameter'][idx_]
        self.projected_detection_statistic['parameter'] = self.projected_detection_statistic['parameter'][idx_]
        

    def find_threshold(self, iteration=10, print_output=True, no_multiprocessing=False):
        """
        Function to find the optimal SNR threshold by maximizing the cross-entropy difference.

        Parameters
        ----------
        iteration : int, optional
            Number of iterations for threshold search. Default is 10.
        print_output : bool, optional
            Whether to print the best SNR threshold. Default is True.

        Returns
        -------
        best_thr : float
            The optimal SNR threshold that maximizes the cross-entropy difference.
        del_H : np.ndarray
            Array of cross-entropy differences for each threshold tested.
        H : np.ndarray
            Array of cross-entropy values for the KDE with cut.
        H_true : np.ndarray
            Array of cross-entropy values for the original KDE.
        snr_thrs : np.ndarray
            Array of SNR thresholds tested.
            
        Raises
        ------
        ValueError
            If the number of iterations is less than 1.

        """

        snr_thrs = np.linspace(
            self.projected_detection_statistic['threshold_search_bounds'][0],
            self.projected_detection_statistic['threshold_search_bounds'][1],
            iteration
        )
        iters = np.arange(iteration)

        sample_size = self.sample_size
        parameters_to_fit = self.parameters_to_fit.copy()
        original_detection_statistic = self.original_detection_statistic.copy()
        projected_detection_statistic = self.projected_detection_statistic.copy()

        # set-up inputs for multoprocessing
        input_args = [(
            snr_thr, 
            sample_size,
            np.array(original_detection_statistic['parameter']),
            np.array(original_detection_statistic['threshold']),
            np.array(projected_detection_statistic['parameter']),
            np.array(parameters_to_fit['parameter']),
            iters[i]
        ) for i, snr_thr in enumerate(snr_thrs)]

        input_args = np.array(input_args, dtype=object)

        # test with for loop first before using multiprocessing
        del_H = np.zeros(iteration)
        H = np.zeros(iteration)
        H_true = np.zeros(iteration)

        if no_multiprocessing:
            for args in tqdm(input_args, total=len(input_args), ncols=100):
                del_H_i, H_i, H_true_i, iter_i = cross_entropy_difference(args)

                

                del_H[iter_i] = del_H_i
                H[iter_i] = H_i
                H_true[iter_i] = H_true_i
        else:
            print("if multiprocessing get stuck, use no_multiprocessing=True")
            npool = self.npool
            with mpp.Pool(processes=npool) as pool:
                self._multiprocessing_error()
                if self.multiprocessing_verbose:
                    for result in tqdm(
                            pool.imap_unordered
                            (cross_entropy_difference, input_args),
                            total=len(input_args),
                            ncols=100,
                        ):
                            del_H_i, H_i, H_true_i, iter_i = result
                            del_H[iter_i] = del_H_i
                            H[iter_i] = H_i
                            H_true[iter_i] = H_true_i
                else:
                    # with map, without tqdm
                    for result in pool.map(cross_entropy_difference, input_args):
                        del_H_i, H_i, H_true_i, iter_i = result
                        del_H[iter_i] = del_H_i
                        H[iter_i] = H_i
                        H_true[iter_i] = H_true_i

        best_thr = self.find_best_SNR_threshold(snr_thrs, del_H)
        if print_output:
            print(f"Best SNR threshold: {best_thr:.2f}")

        return best_thr, del_H, H, H_true, snr_thrs

    def find_best_SNR_threshold(self, thrs, del_H):
        """
        Function to find the best SNR threshold using spline interpolation and optimization.
        
        Parameters
        ----------
        thrs : np.ndarray
            Array of SNR thresholds tested.
        del_H : np.ndarray
            Array of cross-entropy differences for each threshold tested.

        Returns
        -------
        best_thr : float
            The optimal SNR threshold that maximizes the cross-entropy difference.
        """
        
        spline = interp1d(thrs, del_H, kind='cubic')
        min_bound = np.min(thrs)
        max_bound = np.max(thrs)
        best_thr = minimize_scalar(lambda x: -spline(x), bounds=(min_bound, max_bound), method='bounded').x

        return best_thr
    
    def _multiprocessing_error(self):
        """
        Prints an error message when multiprocessing is used.
        """
        # to access multi-cores instead of multithreading
        if mpp.current_process().name != 'MainProcess':
            print(
                "\n\n[ERROR] This multiprocessing code must be run under 'if __name__ == \"__main__\":'.\n"
                "Please wrap your script entry point in this guard.\n"
                "See: https://docs.python.org/3/library/multiprocessing.html#multiprocessing-programming\n"
            )
            raise RuntimeError(
                "\nMultiprocessing code must be run under 'if __name__ == \"__main__\":'.\n\n"
            )


#!/usr/bin/env python
from setuptools import setup, find_packages
import sys
from pathlib import Path
import os

# read the contents of your README file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

# check that python version is 3.10 or above
python_version = sys.version_info
if python_version < (3, 10):
    sys.exit("Python < 3.10 is not supported, aborting setup")

# Read requirements from requirements.txt
def parse_requirements(filename):
    with open(filename, 'r') as f:
        lines = f.read().splitlines()
    return [line.strip() for line in lines if line.strip() and not line.startswith('#')]

requirements = parse_requirements("requirements.txt")

# Get version
version = {}
with open(os.path.join(this_directory, "gwsnr", "_version.py")) as fp:
    exec(fp.read(), version)

setup(
    name='gwsnr',
    version=version['__version__'],
    description='gwsnr: A Python Package for Efficient SNR Calculation of Gravitational Waves',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Hemantakumar Phurailatpam',
    license="MIT",
    author_email='hemantaphurailatpam@gmail.com',
    url='https://github.com/hemantaph/gwsnr',
    packages=find_packages(),
    include_package_data=True,
    package_data={
        'gwsnr': ['ann/ann_data/*', 'core/interpolator_pickle/*'],
        'gwsnr.ann': ['ann_data/*'],
        'gwsnr.core': ['interpolator_pickle/*'],
    },
    python_requires='>=3.10',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    install_requires=requirements,
)



import multiprocessing as mp
# mp.set_start_method('fork', force=True)

from multiprocessing import Pool
import numpy as np
import jax
import jax.numpy as jnp
# jax.config.update("jax_enable_x64", True)
from ripple import ms_to_Mc_eta
from jax import vmap
from jax import jit
from ..jax import findchirp_chirptime_jax

from tqdm import tqdm

from ..utils import noise_weighted_inner_prod_ripple

class RippleInnerProduct:
    """
    Class to compute the noise weighted inner product for a given waveform and PSD
    """

    def __init__(self, 
                 waveform_name='IMRPhenomD', 
                 minimum_frequency=20.0, 
                 sampling_frequency=2048.0, 
                 reference_frequency=None
                ):

        # instance initialization
        self.f_u = sampling_frequency/2
        self.f_ref = reference_frequency if reference_frequency is not None else minimum_frequency
        self.f_l = minimum_frequency

        # get list of arguments for the chosen waveform
        self.arg_list = self.arg_selection(waveform_name) 
        # get the instance to the waveform generator to compute h+ and hx
        # JAX function
        self.gen_hphc = jit(self.select_waveform(waveform_name))
        # use vmap instead of looping
        # JAX function
        self.vmap_waveform = vmap(self.gen_hphc, in_axes=(0, 0, None))
        # use vmap for findchirp chirp time
        # JAX function
        self.vmap_findchirp_chirptime = vmap(findchirp_chirptime_jax, in_axes=(0, 0, None))
        # # use vmap for inner product
        # self.vmap_noise_weighted_inner_product = jit(noise_weighted_inner_product)

        # self.vmap_arange = vmap(jnp.arange, in_axes=(None, 0, 0))

    # def noise_weighted_inner_product(
    #     self, signal1, signal2, psd, duration,
    # ):
    #     """
    #     Noise weighted inner product of two time series data sets.

    #     Parameters
    #     ----------
    #     signal1: `numpy.ndarray` or `float`
    #         First series data set.
    #     signal2: `numpy.ndarray` or `float`
    #         Second series data set.
    #     psd: `numpy.ndarray` or `float`
    #         Power spectral density of the detector.
    #     duration: `float`
    #         Duration of the data.
    #     """

    #     nwip_arr = np.conj(signal1) * signal2 / psd
    #     return 4 / duration * np.sum(nwip_arr)

    def arg_selection(self, waveform_name):
        """
        Returns the list of arguments required for the chosen waveform.

        Parameters
        ----------
        waveform_name: `str`
            The name of the waveform to use. Ripple supported waveforms only.
        
        Returns
        -------
        list: List of arguments required for the chosen waveform.
        """

        supported_waveforms = ['IMRPhenomXAS', 'IMRPhenomD', 'TaylorF2', 'IMRPhenomD_NRTidalv2']

        if waveform_name == 'IMRPhenomXAS':
            return ['Mchirp', 'eta', 'a_1', 'a_2', 'luminosity_distance', 'tc', 'phase', 'theta_jn']
        elif waveform_name == 'IMRPhenomD':
            return ['Mchirp', 'eta', 'a_1', 'a_2', 'luminosity_distance', 'tc', 'phase', 'theta_jn']
        elif waveform_name == 'TaylorF2':
            return ['Mchirp', 'eta', 'a_1', 'a_2', 'lambda_1', 'lambda_2','luminosity_distance', 'tc', 'phase', 'theta_jn']
        elif waveform_name == 'IMRPhenomD_NRTidalv2':
            return ['Mchirp', 'eta', 'a_1', 'a_2', 'lambda_1', 'lambda_2','luminosity_distance', 'tc', 'phase', 'theta_jn']
        else:
            raise ValueError(f"Waveform '{waveform_name}' not supported by ripple. Supported waveforms are: {supported_waveforms}")

    def select_waveform(self, waveform_name):
        """
        Imports and returns the specified waveform from ripple.waveforms.

        Parameters:
        waveform_name (str): The name of the waveform to import.

        Returns:
        class: The waveform class from ripple.waveforms.
        """

        try:
            # Import inside the function
            waveform_module = __import__('ripple.waveforms', fromlist=[waveform_name])
            mod_ = getattr(waveform_module, waveform_name)
            attr_ = "gen_"+waveform_name+"_hphc"
            return getattr(mod_, attr_)
        except AttributeError:
            raise ValueError(f"Waveform '{waveform_name}' not found in ripple.waveforms module.")

    # def waveform_polarization(self, mass_1, mass_2, luminosity_distance, theta_jn, phase, a_1, a_2):

    #     f_u = self.f_u
    #     f_l = self.f_l
    #     f_ref = f_l
    #     duration_max = 64.
    #     duration_min = 4.

    #     # set up waveform calculator inputs
    #     Mchirp, eta = jnp.array(ms_to_Mc_eta(jnp.array([mass_1, mass_2])), dtype=jnp.float32)
    #     tc = 0.
    #     # 'Mchirp', 'eta', 'a_1', 'a_2', 'luminosity_distance', 'tc', 'phase', 'theta_jn'
    #     theta_ripple = jnp.array([Mchirp, eta, a_1, a_2, luminosity_distance, tc, phase, theta_jn]).T

    #     safety = 1.1
    #     approx_duration = safety * self.vmap_findchirp_chirptime(mass_1, mass_2, f_l)  # vmap result give you a stack which is 2D JAX array
    #     duration = jnp.ceil(approx_duration + 2.0).flatten()  # coverts to numpy 1D array
    #     duration = jnp.clip(duration, duration_min, duration_max)

    #     # Frequency array calculation for each of the mass combination
    #     del_f = 1.0 / duration  # Extract scalar from array
    #     f_u = f_u + del_f
    #     fs = self.vmap_arange(f_l, f_u, del_f)

    #     hp, hc = self.gen_hphc(fs, theta_ripple, f_ref)

    #     return hp, hc, fs
        
    
    def noise_weighted_inner_product_jax(self, gw_param_dict, psd_list, detector_list, duration=None, duration_min=2, duration_max=128, npool=4, multiprocessing_verbose=True):
        """
        Compute the noise weighted inner product for a given waveform and PSD.

        Parameters
        ----------
        gw_param_dict: `dict`
            Dictionary containing the waveform parameters. The keys should be the parameter names and the values should be numpy arrays.
        psd_dict: bilby.gw.detector.PowerSpectralDensity object
            Dictionary containing the power spectral density for each detector.
        duration: `float` or `numpy.ndarray`
            Duration of the waveform. 
            Default is None. It will compute the duration based on the chirp time.
        duration_min: `float`
            Minimum duration of the waveform.
            Default is 2s.
        duration_max: `float`
            Maximum duration of the waveform.
            Default is 512s.
        verbose: `bool`
            If True, print the waveform parameters and PSDs.
            Default is False.

        Returns
        -------
        hp_inner_hp: `numpy.ndarray`
            Noise weighted inner product of h+ with h+
        """

        # set up GW parameters
        gw_param_dict = gw_param_dict.copy()
        size = len(gw_param_dict['mass_1'])
        # check dict values
        # for gw_param_dict, values should be an array
        for key, value in gw_param_dict.items():
            if not isinstance(value, np.ndarray):
                raise ValueError(f"Expected numpy array for {key} in gw_param_dict")
            else:
                gw_param_dict[key] = jnp.array(value)

        # set up waveform calculator inputs
        gw_param_dict['Mchirp'],  gw_param_dict['eta'] = ms_to_Mc_eta(jnp.array([ gw_param_dict['mass_1'], gw_param_dict['mass_2'] ]))
        gw_param_dict['tc'] = jnp.zeros(size)

        # set up input arguments for the JAX waveform generator 
        theta_ripple = []
        for key in self.arg_list:
            if key not in gw_param_dict.keys():
                raise ValueError(f"Missing key {key} in gw_param_dict")
            else:
                theta_ripple.append(gw_param_dict[key])
        theta_ripple = jnp.array(theta_ripple).T  # JAX numpy array

        # Set up duration of the signal
        if duration is not None:
            if isinstance(duration, float):
                duration = duration * np.ones(size)
            elif isinstance(duration, np.ndarray):
                if len(duration) != size:
                    raise ValueError(f"Duration array length should be equal to the number of mass combinations")
            else:
                raise ValueError(f"Duration should be either a float or a numpy array")
        else:
            # IMPORTANT: time duration calculation for each of the mass combination
            safety = 1.1
            mass_1 = jnp.array([gw_param_dict['mass_1']])
            mass_2 = jnp.array([gw_param_dict['mass_2']])
            f_l = float(self.f_l)
            approx_duration = safety * self.vmap_findchirp_chirptime(mass_1, mass_2, f_l)  # vmap result give you a stack which is 2D JAX array
            duration = np.ceil(approx_duration + 2.0).flatten()  # coverts to numpy 1D array
            
            if duration_max:
                duration[duration > duration_max] = duration_max  # IMRPheonomXPHM has maximum duration of 371s
            if duration_min:
                duration[duration < duration_min] = duration_min

        # Frequency array calculation for each of the mass combination
        del_f = 1.0 / duration
        fs = []
        # Frequency array size should be same for all events 
        # padding np.nan elements will be used to achieve this
        fsize_max = int((self.f_u - 0.0) / min(del_f)) + 2  # +2 to include upper limit
        fsize_arr = []
        for df  in del_f:
            # NOTE: frequency array start from 0Hz but h+,hx will be set to zero below fmin or self.f_l  
            flist = np.arange(0.0, self.f_u + df, df)
            fsize_arr.append(len(flist))  # record the array size to know how much padding will be added
            # add np.nan padding elements 
            if len(flist) <= fsize_max:
                # Add NaN values to make the array size equal to fsize_max
                flist = np.concatenate([flist, np.full(fsize_max - len(flist), np.nan)])
            fs.append(flist)

        # Convert list of arrays to a 2D JAX array
        fs = jnp.array(fs)

        # set up reference_frequency for vmap input
        f_ref = float(self.f_ref)
        # compute the waveform h+ and hx
        # NOTE: result will be in complex64 instead of complex128. This will be change later
        # vmap+jax.jit faster than just using jax.jit
        # mp.set_start_method('spawn', force=True)
        hp, hc = self.vmap_waveform(fs, theta_ripple, f_ref)
        hp, hc = np.array(hp, dtype=np.complex128), np.array(hc, dtype=np.complex128)
        # print(f"fs : {fs}")
        # print(f"theta_ripple : {theta_ripple}")
        # print(f"hp : {hp}")
        # print(f"hc : {hc}")
        fs = np.array(fs, dtype=np.float64)

        input_arguments = [
            (hp_i, hc_i, fs_i, fsize_arr_i, self.f_l, duration_i, idx, psd_list)
            for idx, (hp_i, hc_i, fs_i, fsize_arr_i, duration_i)
                in enumerate(zip(hp, hc, fs, fsize_arr, duration))
        ]

        num_det = len(detector_list)
        # np.shape(hp_inner_hp) = (len(num_det), size1)
        hp_inner_hp = np.zeros((num_det, size), dtype=np.complex128)
        hc_inner_hc = np.zeros((num_det, size), dtype=np.complex128)

        # for j in range(size):
        #     result = noise_weighted_inner_prod_ripple(input_arguments[j])
        #     hp_inner_hp_i, hc_inner_hc_i, iter_i = result
        #     hp_inner_hp[:, iter_i] = hp_inner_hp_i
        #     hc_inner_hc[:, iter_i] = hc_inner_hc_i

        # mp.set_start_method('fork', force=True)

        with Pool(processes=npool) as pool:
            # call the same function with different data in parallel
            # imap->retain order in the list, while map->doesn't
            if multiprocessing_verbose:
                for result in tqdm(
                    pool.imap_unordered(noise_weighted_inner_prod_ripple, input_arguments),
                    total=len(input_arguments),
                    ncols=100,
                ):
                    # but, np.shape(hp_inner_hp_i) = (size1, len(num_det))
                    hp_inner_hp_i, hc_inner_hc_i, iter_i = result
                    hp_inner_hp[:, iter_i] = hp_inner_hp_i
                    hc_inner_hc[:, iter_i] = hc_inner_hc_i
            else:
                # with map, without tqdm
                for result in pool.map(noise_weighted_inner_prod_ripple, input_arguments):
                    hp_inner_hp_i, hc_inner_hc_i, iter_i = result
                    hp_inner_hp[:, iter_i] = hp_inner_hp_i
                    hc_inner_hc[:, iter_i] = hc_inner_hc_i
        # mp.set_start_method('spawn', force=True)

        return hp_inner_hp, hc_inner_hc

    # def noise_weighted_inner_product_jax(self, gw_param_dict, psd_list, detector_list, duration=None, duration_min=2, duration_max=128, npool=4, multiprocessing_verbose=True):
    #     """
    #     Compute the noise weighted inner product for a given waveform and PSD.

    #     Parameters
    #     ----------
    #     gw_param_dict: `dict`
    #         Dictionary containing the waveform parameters. The keys should be the parameter names and the values should be numpy arrays.
    #     psd_dict: bilby.gw.detector.PowerSpectralDensity object
    #         Dictionary containing the power spectral density for each detector.
    #     duration: `float` or `numpy.ndarray`
    #         Duration of the waveform. 
    #         Default is None. It will compute the duration based on the chirp time.
    #     duration_min: `float`
    #         Minimum duration of the waveform.
    #         Default is 2s.
    #     duration_max: `float`
    #         Maximum duration of the waveform.
    #         Default is 512s.
    #     verbose: `bool`
    #         If True, print the waveform parameters and PSDs.
    #         Default is False.

    #     Returns
    #     -------
    #     hp_inner_hp: `numpy.ndarray`
    #         Noise weighted inner product of h+ with h+
    #     """

    #     # set up GW parameters
    #     gw_param_dict = gw_param_dict.copy()
    #     size = len(gw_param_dict['mass_1'])
    #     # check dict values
    #     # for gw_param_dict, values should be an array
    #     for key, value in gw_param_dict.items():
    #         if not isinstance(value, np.ndarray):
    #             raise ValueError(f"Expected numpy array for {key} in gw_param_dict")
    #         else:
    #             gw_param_dict[key] = jnp.array(value)

    #     # set up waveform calculator inputs
    #     gw_param_dict['Mchirp'],  gw_param_dict['eta'] = ms_to_Mc_eta(jnp.array([ gw_param_dict['mass_1'], gw_param_dict['mass_2'] ]))
    #     gw_param_dict['tc'] = jnp.zeros(size)

    #     # set up input arguments for the JAX waveform generator 
    #     theta_ripple = []
    #     for key in self.arg_list:
    #         if key not in gw_param_dict.keys():
    #             raise ValueError(f"Missing key {key} in gw_param_dict")
    #         else:
    #             theta_ripple.append(gw_param_dict[key])
    #     theta_ripple = jnp.array(theta_ripple).T  # JAX numpy array

    #     # Set up duration of the signal
    #     if duration is not None:
    #         if isinstance(duration, float):
    #             duration = duration * np.ones(size)
    #         elif isinstance(duration, np.ndarray):
    #             if len(duration) != size:
    #                 raise ValueError(f"Duration array length should be equal to the number of mass combinations")
    #         else:
    #             raise ValueError(f"Duration should be either a float or a numpy array")
    #     else:
    #         # IMPORTANT: time duration calculation for each of the mass combination
    #         safety = 1.1
    #         mass_1 = jnp.array([gw_param_dict['mass_1']]).T
    #         mass_2 = jnp.array([gw_param_dict['mass_2']]).T
    #         f_l = self.f_l*jnp.ones((size,1))
    #         approx_duration = safety * self.vmap_findchirp_chirptime(mass_1, mass_2, f_l)  # vmap result give you a stack which is 2D JAX array
    #         duration = np.ceil(approx_duration + 2.0).flatten()  # coverts to numpy 1D array
            
    #         if duration_max:
    #             duration[duration > duration_max] = duration_max  # IMRPheonomXPHM has maximum duration of 371s
    #         if duration_min:
    #             duration[duration < duration_min] = duration_min

    #     # Frequency array calculation for each of the mass combination
    #     del_f = 1.0 / duration
    #     fs = []
    #     # Frequency array size should be same for all events 
    #     # padding np.nan elements will be used to achieve this
    #     fsize_max = int((self.f_u - 0.0) / min(del_f)) + 2  # +2 to include upper limit
    #     fsize_arr = []
    #     for df  in del_f:
    #         # NOTE: frequency array start from 0Hz but h+,hx will be set to zero below fmin or self.f_l  
    #         flist = np.arange(0.0, self.f_u + df, df)
    #         fsize_arr.append(len(flist))  # record the array size to know how much padding will be added
    #         # add np.nan padding elements 
    #         if len(flist) <= fsize_max:
    #             # Add NaN values to make the array size equal to fsize_max
    #             flist = np.concatenate([flist, np.full(fsize_max - len(flist), np.nan)])
    #         fs.append(flist)
    #     # Convert list of arrays to a 2D JAX array
    #     fs = jnp.array(fs)

    #     # set up reference_frequency for vmap input
    #     f_ref = self.f_ref * jnp.ones((size,1))
    #     # compute the waveform h+ and hx
    #     # NOTE: result will be in complex64 instead of complex128. This will be change later
    #     # vmap+jax.jit faster than just using jax.jit
    #     # mp.set_start_method('spawn', force=True)
    #     hp, hc = self.vmap_waveform(fs, theta_ripple, f_ref)
    #     hp, hc = np.array(hp, dtype=np.complex128), np.array(hc, dtype=np.complex128)
    #     fs = np.array(fs, dtype=np.float64)

    #     input_arguments = []
    #     for i in range(size):
    #         input_arguments.append(
    #             [
    #                 hp[i],
    #                 hc[i],
    #                 fs[i],
    #                 fsize_arr[i],
    #                 self.f_l,
    #                 duration[i],
    #                 i,
    #                 psd_list,
    #             ]
    #         )
    #     input_arguments = np.array(input_arguments, dtype=object)

    #     num_det = len(detector_list)
    #     # np.shape(hp_inner_hp) = (len(num_det), size1)
    #     hp_inner_hp = np.zeros((num_det, size), dtype=np.complex128)
    #     hc_inner_hc = np.zeros((num_det, size), dtype=np.complex128)

    #     # for j in range(size):
    #     #     result = noise_weighted_inner_prod_ripple(input_arguments[j])
    #     #     hp_inner_hp_i, hc_inner_hc_i, iter_i = result
    #     #     hp_inner_hp[:, iter_i] = hp_inner_hp_i
    #     #     hc_inner_hc[:, iter_i] = hc_inner_hc_i

    #     # mp.set_start_method('fork', force=True)

    #     with Pool(processes=npool) as pool:
    #         # call the same function with different data in parallel
    #         # imap->retain order in the list, while map->doesn't
    #         if multiprocessing_verbose:
    #             for result in tqdm(
    #                 pool.imap_unordered(noise_weighted_inner_prod_ripple, input_arguments),
    #                 total=len(input_arguments),
    #                 ncols=100,
    #             ):
    #                 # but, np.shape(hp_inner_hp_i) = (size1, len(num_det))
    #                 hp_inner_hp_i, hc_inner_hc_i, iter_i = result
    #                 hp_inner_hp[:, iter_i] = hp_inner_hp_i
    #                 hc_inner_hc[:, iter_i] = hc_inner_hc_i
    #         else:
    #             # with map, without tqdm
    #             for result in pool.map(noise_weighted_inner_prod_ripple, input_arguments):
    #                 hp_inner_hp_i, hc_inner_hc_i, iter_i = result
    #                 hp_inner_hp[:, iter_i] = hp_inner_hp_i
    #                 hc_inner_hc[:, iter_i] = hc_inner_hc_i
    #     # mp.set_start_method('spawn', force=True)

    #     return hp_inner_hp, hc_inner_hc
        

        # return input_arguments


        # # set up psd
        # psd_dict = psd_dict.copy()
        # for key, value in psd_dict.items(): # key is the detector name
        #     if not isinstance(value, bilby.gw.detector.PowerSpectralDensity):
        #         raise ValueError(f"Expected bilby.gw.detector.PowerSpectralDensity object for {key} in psd_dict")
        #     # compute the power spectral density array according to the frequency array
        #     psd_list = []
        #     fs_ = []
        #     for i in range(size):
        #         fs_.append(fs[i][:fsize_arr[i]])  # avoid np.nan paddings
        #         psd_list.append(psd_dict[key].get_power_spectral_density_array(fs_[i]))
        #     psd_dict[key] = np.array(psd_list, dtype=object)
        #     fs = np.array(fs_, dtype=object)
    
        # # set up h+,hx for inner_product
        # hp_, hc_ = [], []
        # for i in range(size):
        #     # remove the np.nan padding
        #     hp_.append(np.array(hp[i][:fsize_arr[i]], dtype=np.complex128))
        #     hc_.append(np.array(hc[i][:fsize_arr[i]], dtype=np.complex128))
        #     # find the index of 20Hz or nearby
        #     # set all elements to zero below this index
        #     idx = np.abs(fs[i] - self.f_l).argmin()
        #     hp_[i][0:idx] = 0.0 + 0.0j
        #     hc_[i][0:idx] = 0.0 + 0.0j
        # # each row don't have the same length, so keep the array type to object
        # hp = np.array(hp_, dtype=object)
        # hc = np.array(hc_, dtype=object)
        # #del hp_, hc_  # free up memory


        # # compute the noise weighted inner product
        # hp_inner_hp_list = []
        # hc_inner_hc_list = []
        # for det, psd_list in psd_dict.items():
        #     hp_inner_hp = []
        #     hc_inner_hc = []
        #     for i in range(size):  # loop over parameters
        #         psd_ = np.array(psd_list[i])
        #         duration_ = duration[i]
        #         idx2 = (psd_ != 0.0) & (psd_ != np.inf)  # this is necessary to avoid problem in np.sum 
        #         # it's cumbersome to use jax jitting noise_weighted_inner_product
        #         # so, I will use numba jitted one instead
        #         hp_inner_hp.append(self.vmap_noise_weighted_inner_product(
        #             hp[i][idx2],
        #             hp[i][idx2],
        #             psd_[idx2],
        #             duration_,
        #         ))

        #         hc_inner_hc.append(noise_weighted_inner_product(
        #             hc[i][idx2],
        #             hc[i][idx2],
        #             psd_[idx2],
        #             duration_,
        #         ))

        #     hp_inner_hp_list.append(hp_inner_hp)
        #     hc_inner_hc_list.append(hc_inner_hc)

        # hp_inner_hp = np.array(hp_inner_hp_list)
        # hc_inner_hc = np.array(hc_inner_hc_list)



        # # del psd_dict
        # # del gw_param_dict

        # return hp_inner_hp, hc_inner_hc


from .mlx_functions import (
    findchirp_chirptime_mlx,
    antenna_response_plus,
    antenna_response_cross,
    antenna_response_array,
)

from .mlx_interpolators import (
    get_interpolated_snr_aligned_spins_mlx,
    get_interpolated_snr_no_spins_mlx,
)

# __all__ = [
#     # mlx_functions
#     'findchirp_chirptime_mlx',
#     'antenna_response_plus',
#     'antenna_response_cross',
#     'antenna_response_array',
#     # mlx_interpolators
#     'get_interpolated_snr_aligned_spins_mlx',
#     'get_interpolated_snr_no_spins_mlx',
# ]


# -*- coding: utf-8 -*-
"""
Numba-compiled helper functions for gravitational wave signal-to-noise ratio calculations.

This module provides optimized numerical functions for gravitational wave data analysis,
including chirp time calculations, antenna response computations, polarization tensors,
coordinate transformations, and noise-weighted inner products. All functions are compiled
with Numba's @njit decorator for high-performance computation, with parallel processing
support using prange for multi-threaded execution where applicable.
"""

# -*- coding: utf-8 -*-
"""
Helper functions for gwsnr. All functions are njit compiled.
"""

import numpy as np
from numba import njit, prange

Gamma = 0.5772156649015329
Pi = np.pi
MTSUN_SI = 4.925491025543576e-06

@njit
def findchirp_chirptime(m1, m2, fmin):
    """
    Time taken from f_min to f_lso (last stable orbit). 3.5PN in fourier phase considered.

    Parameters
    ----------
    m1 : `float`
        Mass of the first body in solar masses.
    m2 : `float`
        Mass of the second body in solar masses.
    fmin : `float`
        Lower frequency cutoff.

    Returns
    -------
    chirp_time : float
        Time taken from f_min to f_lso (last stable orbit frequency).
    """

    # variables used to compute chirp time
    m = m1 + m2
    eta = m1 * m2 / m / m
    c0T = c2T = c3T = c4T = c5T = c6T = c6LogT = c7T = 0.0

    c7T = Pi * (
        14809.0 * eta * eta / 378.0 - 75703.0 * eta / 756.0 - 15419335.0 / 127008.0
    )

    c6T = (
        Gamma * 6848.0 / 105.0
        - 10052469856691.0 / 23471078400.0
        + Pi * Pi * 128.0 / 3.0
        + eta * (3147553127.0 / 3048192.0 - Pi * Pi * 451.0 / 12.0)
        - eta * eta * 15211.0 / 1728.0
        + eta * eta * eta * 25565.0 / 1296.0
        + eta * eta * eta * 25565.0 / 1296.0
        + np.log(4.0) * 6848.0 / 105.0
    )
    c6LogT = 6848.0 / 105.0

    c5T = 13.0 * Pi * eta / 3.0 - 7729.0 * Pi / 252.0

    c4T = 3058673.0 / 508032.0 + eta * (5429.0 / 504.0 + eta * 617.0 / 72.0)
    c3T = -32.0 * Pi / 5.0
    c2T = 743.0 / 252.0 + eta * 11.0 / 3.0
    c0T = 5.0 * m * MTSUN_SI / (256.0 * eta)

    # This is the PN parameter v evaluated at the lower freq. cutoff
    xT = np.power(Pi * m * MTSUN_SI * fmin, 1.0 / 3.0)
    x2T = xT * xT
    x3T = xT * x2T
    x4T = x2T * x2T
    x5T = x2T * x3T
    x6T = x3T * x3T
    x7T = x3T * x4T
    x8T = x4T * x4T

    # Computes the chirp time as tC = t(v_low)
    # tC = t(v_low) - t(v_upper) would be more
    # correct, but the difference is negligble.
    return (
        c0T
        * (
            1
            + c2T * x2T
            + c3T * x3T
            + c4T * x4T
            + c5T * x5T
            + (c6T + c6LogT * np.log(xT)) * x6T
            + c7T * x7T
        )
        / x8T
    )

@njit
def einsum1(m,n):
    """
    Function to calculate einsum of two 3x1 vectors

    Parameters
    ----------
    m : `numpy.ndarray`
        3x1 vector.
    n : `numpy.ndarray`
        3x1 vector.
        
    Returns
    -------
    ans : `numpy.ndarray`
        3x3 matrix.
    """
    ans = np.zeros((3,3))
    ans[0,0] = m[0]*n[0]
    ans[0,1] = m[0]*n[1]
    ans[0,2] = m[0]*n[2]
    ans[1,0] = m[1]*n[0]
    ans[1,1] = m[1]*n[1]
    ans[1,2] = m[1]*n[2]
    ans[2,0] = m[2]*n[0]
    ans[2,1] = m[2]*n[1]
    ans[2,2] = m[2]*n[2]
    return ans

@njit
def einsum2(m,n):
    """
    Function to calculate einsum of two 3x3 matrices

    Parameters
    ----------
    m : `numpy.ndarray`
        3x3 matrix.
    n : `numpy.ndarray`
        3x3 matrix.

    Returns
    -------
    ans : `numpy.ndarray`
        3x3 matrix.
    """
    ans = m[0,0]*n[0,0] + m[0,1]*n[0,1] + m[0,2]*n[0,2] + m[1,0]*n[1,0] + m[1,1]*n[1,1] + m[1,2]*n[1,2] + m[2,0]*n[2,0] + m[2,1]*n[2,1] + m[2,2]*n[2,2]
    return ans

@njit
def gps_to_gmst(gps_time):
    """
    Function to convert gps time to greenwich mean sidereal time

    Parameters
    ----------
    gps_time : `float`
        GPS time in seconds.

    Returns
    -------
    gmst : `float`
        Greenwich mean sidereal time in radians.
    """
    slope = 7.292115855425873e-05
    intercept = -45991.08966925838
    return slope*gps_time+intercept

@njit
def ra_dec_to_theta_phi(ra, dec, gmst):
    """
    Function to convert ra and dec to theta and phi

    Parameters
    ----------
    ra : `float`
        Right ascension of the source in radians.
    dec : `float`
        Declination of the source in radians.
    gmst : `float`
        Greenwich mean sidereal time in radians.

    Returns
    -------
    theta : `float`
        Polar angle in radians.
    phi : `float`
        Azimuthal angle in radians.
    """

    phi = ra - gmst
    theta = np.pi / 2.0 - dec
    return theta, phi

@njit
def get_polarization_tensor_plus(ra, dec, time, psi):
    """
    Function to calculate the polarization tensor

    Parameters
    ----------
    ra : `float`
        Right ascension of the source in radians.
    dec : float
        Declination of the source in radians.
    time : `float`
        GPS time of the source.
    psi : `float`
        Polarization angle of the source.

    Returns
    -------
    polarization_tensor: `numpy.ndarray`
        Polarization tensor of the detector.
    """
    gmst = np.fmod(gps_to_gmst(time), 2 * np.pi)
    theta, phi = ra_dec_to_theta_phi(ra, dec, gmst)
    u = np.array([np.cos(phi) * np.cos(theta), np.cos(theta) * np.sin(phi), -np.sin(theta)])
    v = np.array([-np.sin(phi), np.cos(phi), 0])
    m = -u * np.sin(psi) - v * np.cos(psi)
    n = -u * np.cos(psi) + v * np.sin(psi)

    return einsum1(m, m) - einsum1(n, n)
    
@njit
def get_polarization_tensor_cross(ra, dec, time, psi):
    """
    Function to calculate the polarization tensor

    Parameters
    ----------
    ra : `float`
        Right ascension of the source in radians.
    dec : float
        Declination of the source in radians.
    time : `float`
        GPS time of the source.
    psi : `float`
        Polarization angle of the source.

    Returns
    -------
    polarization_tensor: `numpy.ndarray`
        Polarization tensor of the detector.
    """
    gmst = np.fmod(gps_to_gmst(time), 2 * np.pi)
    theta, phi = ra_dec_to_theta_phi(ra, dec, gmst)
    u = np.array([np.cos(phi) * np.cos(theta), np.cos(theta) * np.sin(phi), -np.sin(theta)])
    v = np.array([-np.sin(phi), np.cos(phi), 0])
    m = -u * np.sin(psi) - v * np.cos(psi)
    n = -u * np.cos(psi) + v * np.sin(psi)

    return einsum1(m, n) + einsum1(n, m)

@njit
def antenna_response_plus(ra, dec, time, psi, detector_tensor):
    """
    Function to calculate the antenna response

    Parameters
    ----------
    ra : `float`
        Right ascension of the source in radians.
    dec : float
        Declination of the source in radians.
    time : `float`
        GPS time of the source.
    psi : `float`
        Polarization angle of the source.
    detector_tensor : array-like
        Detector tensor for the detector (3x3 matrix)
    mode : `str`
        Mode of the polarization. Default is 'plus'.

    Returns
    -------
    antenna_response: `float`
        Antenna response of the detector.
    """

    polarization_tensor = get_polarization_tensor_plus(ra, dec, time, psi)
    return einsum2(detector_tensor, polarization_tensor)

@njit
def antenna_response_cross(ra, dec, time, psi, detector_tensor):
    """
    Function to calculate the antenna response

    Parameters
    ----------
    ra : `float`
        Right ascension of the source in radians.
    dec : float
        Declination of the source in radians.
    time : `float`
        GPS time of the source.
    psi : `float`
        Polarization angle of the source.
    detector_tensor : array-like
        Detector tensor for the detector (3x3 matrix)
    mode : `str`
        Mode of the polarization. Default is 'plus'.

    Returns
    -------
    antenna_response: `float`
        Antenna response of the detector.
    """

    polarization_tensor = get_polarization_tensor_cross(ra, dec, time, psi)
    return einsum2(detector_tensor, polarization_tensor)

@njit(parallel=True)
def antenna_response_array(ra, dec, time, psi, detector_tensor):
    """
    Function to calculate the antenna response in array form.

    Parameters
    ----------
    ra : `numpy.ndarray`
        Right ascension of the source in radians.
    dec : `numpy.ndarray`
        Declination of the source in radians.
    time : `numpy.ndarray`
        GPS time of the source.
    psi : `numpy.ndarray`
        Polarization angle of the source.
    detector_tensor : array-like
        Detector tensor for the multiple detectors (nx3x3 matrix), where n is the number of detectors.

    Returns
    -------
    antenna_response: `numpy.ndarray`
        Antenna response of the detector. Shape is (n, len(ra)).
    """

    len_det = len(detector_tensor)
    len_param = len(ra) 
    Fp = np.zeros((len_det, len_param))
    Fc = np.zeros((len_det, len_param))

    for i in prange(len_param):
        for j in range(len_det):
        
            Fp[j,i] = antenna_response_plus(ra[i], dec[i], time[i], psi[i], detector_tensor[j])
            Fc[j,i] = antenna_response_cross(ra[i], dec[i], time[i], psi[i], detector_tensor[j])

    return Fp, Fc

@njit
def effective_distance(
    luminosity_distance, theta_jn, ra, dec, geocent_time, psi, detector_tensor
):
    """
    Function to calculate the effective distance of the source.

    Parameters
    ----------
    luminosity_distance : `float`
        Luminosity distance of the source in Mpc.
    theta_jn : `float`
        Angle between the line of sight and the orbital angular momentum vector.
    ra : `float`
        Right ascension of the source in radians.
    dec : `float`
        Declination of the source in radians.
    time : `float`
        GPS time of the source.
    psi : `float`
        Polarization angle of the source.
    detector_tensor : array-like
        Detector tensor for the detector (3x3 matrix).

    Returns
    -------
    effective_distance: `float`
        Effective distance of the source in Mpc.
    """

    Fp, Fc = antenna_response_plus(ra, dec, geocent_time, psi, detector_tensor), antenna_response_cross(ra, dec, geocent_time, psi, detector_tensor)

    return luminosity_distance / np.sqrt(
            Fp**2 * ((1 + np.cos(theta_jn) ** 2) / 2) ** 2
            + Fc**2 * np.cos(theta_jn) ** 2
        )

@njit(parallel=True)
def effective_distance_array(
    luminosity_distance, theta_jn, ra, dec, geocent_time, psi, detector_tensor
):
    """
    Function to calculate the effective distance of the source in array form.

    Parameters
    ----------
    luminosity_distance : `numpy.ndarray`
        Luminosity distance of the source in Mpc.
    theta_jn : `numpy.ndarray`
        Angle between the line of sight and the orbital angular momentum vector.
    ra : `numpy.ndarray`
        Right ascension of the source in radians.
    dec : `numpy.ndarray`
        Declination of the source in radians.
    time : `numpy.ndarray`
        GPS time of the source.
    psi : `numpy.ndarray`
        Polarization angle of the source.
    detector_tensor : array-like
        Detector tensor for the multiple detectors (nx3x3 matrix), where n is the number of detectors.

    Returns
    -------
    effective_distance: `numpy.ndarray`
        Effective distance of the source in Mpc. Shape is (n, len(ra)).
    """

    len_det = len(detector_tensor)
    len_param = len(ra) 
    eff_dist = np.zeros((len_det, len_param))

    for i in prange(len_param):
        for j in range(len_det):
            eff_dist[j,i] = effective_distance(
                luminosity_distance[i], theta_jn[i], ra[i], dec[i], geocent_time[i], psi[i], detector_tensor[j]
            )

    return eff_dist

@njit
def noise_weighted_inner_product(
    signal1, signal2, psd, duration,
):
    """
    Noise weighted inner product of two time series data sets.

    Parameters
    ----------
    signal1: `numpy.ndarray` or `float`
        First series data set.
    signal2: `numpy.ndarray` or `float`
        Second series data set.
    psd: `numpy.ndarray` or `float`
        Power spectral density of the detector.
    duration: `float`
        Duration of the data.
    """

    nwip_arr = np.conj(signal1) * signal2 / psd
    return 4 / duration * np.sum(nwip_arr)

@njit(parallel=True)
def linear_interpolator(xnew_array, y_array, x_array, fill_value=np.inf):
    """
    Linear interpolator for 1D data.

    Parameters
    ----------
    xnew_array : `numpy.ndarray`
        New x values to interpolate.
    y_array : `numpy.ndarray`
        y values corresponding to the x_array.
    x_array : `numpy.ndarray`
        Original x values.

    Returns
    -------
    result : `numpy.ndarray`
        Interpolated y values at xnew_array.
    """

    result = np.zeros_like(xnew_array)
    len_ = xnew_array.shape[0]

    for j in prange(len_):
        xnew = xnew_array[j]
        # Handling extrapolation
        i = np.searchsorted(x_array, xnew) - 1
        # Linear interpolation
        if (i < len(x_array) - 1) and (i > 0):
            x0, x1 = x_array[i], x_array[i + 1]
            y0, y1 = y_array[i], y_array[i + 1]
            result[j] = y0 + (y1 - y0) * (xnew - x0) / (x1 - x0)
        else:
            result[j] = fill_value

    return result

# @njit
# def _helper_hphc(hp,hc,fsize_arr,fs,size,f_l,i):
#     # remove the np.nan padding
#     hp_ = np.array(hp[i][:fsize_arr[i]], dtype=np.complex128)
#     hc_ = np.array(hc[i][:fsize_arr[i]], dtype=np.complex128)
#     # find the index of 20Hz or nearby
#     # set all elements to zero below this index
#     idx = np.abs(fs[i] - f_l).argmin()
#     hp_[i][0:idx] = 0.0 + 0.0j
#     hc_[i][0:idx] = 0.0 + 0.0j

#     return hp_,hc_
